################
## New S4 class to represent list of HTCexp objects
################

## class definition
setClass("HTClist",
         prototype = prototype(elementType = "HTCexp"),
         contains = "list")


setValidity("HTClist",
            function(object){
                fails <- character(0)
                
                ## Check Chromosome Pairs Data
                 chrom <- sapply(object, function(x){
                     paste(seqlevels(x_intervals(x)), seqlevels(y_intervals(x)), sep="")
                 })
                 if (any(duplicated(chrom))){
                     fails <- c(fails, "Multiple 'HTCexp' found from the same chromosome pairs")
                 }

                ## Check Chromosome bounderies
                ##if (length(unique(unlist(ranges(object)))) > length(seqlevels(object))){
                ##    fails <- c(fails, "Same chromosome found with different ranges")
                ##}
                
                if (length(fails) > 0) return(fails)
                return(TRUE)
            })

## constructor
HTClist <- function(...)
{
  listData <- list(...)
  stopifnot(length(listData) > 0L)
  
  if (length(listData) == 1L && is.list(listData[[1L]]))
      listData <- listData[[1L]]
  if (!all(sapply(listData, is, "HTCexp")))
      stop("all elements in '...' must be HTCexp objects")
  listData <- unname(listData)

  names(listData) <- sapply(listData, function(x){
    paste(seqlevels(y_intervals(x)), seqlevels(x_intervals(x)), sep="")
  })
  
  new("HTClist", listData)
}


################
##
## Methods
##
################

setMethod("[", "HTClist",
    function(x, i, ...)
    {
        if (is.character(i)){
            i <- match( i, names(x))
          }        
        HTClist(unlist(x)[i])
    })          


setMethod("as.list", "HTClist",
    function(x)
    {
        as.list(unlist(x))
    })          


setMethod("c", "HTClist", function(x, ...){
    if (missing(x))
        args <- unname(list(...))
    else{
        l <- list(...)
        l <- sapply(l, function(x){
            if (inherits(x,"HTClist"))
                x <- as.list(x)
            x
        })
        args <- unname(c(as.list(x), l))
    }
    HTClist(args)
})


setMethod("detail",signature(x="HTClist"),
          function(x){
              stopifnot(validObject(x))
              lapply(x, show)
              invisible(NULL)
          })


setMethod(f="forcePairwise", signature(x="HTClist"),
          function(x){
              stopifnot(isComplete(x))
              chrs <- seqlevels(x)
              pchr <- pair.chrom(chrs)
              isin <- rep(0, length(pchr))
              names(isin) <- names(pchr)
              isin[names(x)] <- 1
              
              ptoadd <- pchr[names(which(isin==0))]
              
              nmaps <- mclapply(ptoadd, function(obj){
                  symobj <- x[[paste0(obj[2], obj[1])]]
                  HTCexp(intdata=t(intdata(symobj)), xgi=y_intervals(symobj), ygi=x_intervals(symobj))
              })
              c(x, nmaps)
          })

setMethod(f="forceSymmetric", signature(x="HTClist", uplo="missing"),
          function(x, uplo){forceSymmetric(x, uplo="U")})

setMethod(f="forceSymmetric", signature(x="HTClist", uplo="character"),
          function(x, uplo){
              stopifnot(uplo=="U" || uplo=="L")
              stopifnot(isComplete(x))
              x <- sortSeqlevels(x)
              chrs <- seqlevels(x)
              pch <- pair.chrom(chrs, use.order=FALSE)
              if (uplo=="L"){
                  pch <- lapply(pch, "[", 2:1)
                  names(pch) <- sapply(pch, paste0, collapse="")
              }
              x[names(pch)]
          })


setMethod(f="isPairwise", signature(x="HTClist"),
          function(x){
              chrs <- seqlevels(x)
              all.pairs <- names(pair.chrom(chrs))
              ret <- FALSE
              if (length(setdiff(names(pair.chrom(chrs)), names(x)))==0)
                  ret <- TRUE
              ret
          })


setMethod(f="getCombinedIntervals", signature(x="HTClist"),
          function(x){
              x <- x[isIntraChrom(x)]
              ygr <- lapply(x, y_intervals)
              ygi <- suppressWarnings(do.call(c,unname(ygr)))
              
              if(length(which(!sapply(x, isSymmetric)))>0){
                  xgr <- lapply(x, x_intervals)
                  xgi <- suppressWarnings(do.call(c,unname(xgr)))
              }else{
                  xgi <- NULL
              }
              list(ygi=ygi, xgi=xgi)
          })


setMethod(f="getCombinedContacts", signature(x="HTClist"),
          function(x){
              stopifnot(isComplete(x))
              if(!isPairwise(x))
                  x <- forcePairwise(x)
              
              message("Start combining HTCexp objects ...")
              pchr <- t(as.data.frame(pair.chrom(seqlevels(x))))
              dimchr <- t(sapply(x[isIntraChrom(x)], function(xx){dim(intdata(xx))}))
              rownames(dimchr) <- pchr[rownames(dimchr),1]
              col.merged <- lapply(seqlevels(x), function(chr){
                  allpairs <- rownames(pchr)[which(pchr[,1]==chr)]
                  do.call(cBind,lapply(x[allpairs], function(xx){as(intdata(xx), "sparseMatrix")}))
              })
              bigMat <- do.call(rBind, col.merged)
              combint <- getCombinedIntervals(x)
              ##ym <- paste0(paste0("bin",1:length(combint$ygi)),"|",seqnames(combint$ygi),":",start(combint$ygi),"-",end(combint$ygi))
              ym <- id(combint$ygi)
              if (!is.null(combint$xgi)){
                  ##xm <- paste0(paste0("bin",1:length(combint$xgi)),"|",seqnames(combint$xgi),":",start(combint$xgi),"-",end(combint$xgi))
                  xm <- id(combint$xgi)
              }else{
                  xm <- ym
              }
              colnames(bigMat) <- xm
              rownames(bigMat) <- ym
              message("Object size: ",object.size(bigMat))
              bigMat
          })


setMethod(f="isComplete", signature(x="HTClist"),
          function(x){
              chrs <- seqlevels(x)
              lchrs <- length(chrs)
              mat.pairs <- matrix(0, ncol=lchrs, nrow=lchrs, dimnames=list(chrs, chrs))
              for (m in x){
                  mat.pairs[seqlevels(m@ygi), seqlevels(m@xgi)] <- 1
              }         
              ret <- TRUE
              maps_per_chr <- apply(mat.pairs, 1, sum) + apply(mat.pairs, 2, sum)
              if (any(maps_per_chr<(length(seqlevels(x))+1)))
                  ret <- FALSE
              ret
          })


setMethod(f="isBinned", signature(x="HTClist"),
          function(x){
              sapply(x, isBinned)
          })


setMethod(f="isIntraChrom", signature(x="HTClist"),
          function(x){
              sapply(x, isIntraChrom) 
          })


setMethod(f="ranges", signature(x="HTClist"),
          function(x){
              GRangesList(sapply(x, range)) 
          })


setMethod(f="range", signature(x="HTClist"),
          function(x){
              reduce(unlist(ranges(x)))
        })


setMethod(f="seqlevels", signature(x="HTClist"),
          function(x){
              unique(unlist(unname(sapply(x, seqlevels))))
          })


setMethod("show",signature="HTClist",
          function(object){
              stopifnot(validObject(object))
              cat("HTClist object of length",length(object),"\n")
              im <- isIntraChrom(object)
              cat(length(which(im)),"intra /", length(which(!im)), "inter-chromosomal maps\n")
              invisible(NULL)
          })


setMethod("sortSeqlevels", signature="HTClist",
          function(x){
              stopifnot(validObject(x))
              chrs <- seqlevels(x)
              chrs.sort <- sortSeqlevels(chrs)
              out <- x[intersect(names(pair.chrom(chrs.sort)), names(x))]
              out
          })


setMethod("summary", signature=c(object="HTClist"),
           function(object){
               sy <- as.data.frame(t(as.data.frame(lapply(object, summary))))
               rownames(sy) <- paste0(sy$seq1, sy$seq2)
               sy
           })

